/* j2kaudio.cpp

j2kaudio License 200707231230

Copyright (c) 2007 Gregory Alan Hildstrom

All rights reserved.

Permission is hereby granted, free of charge, to any person (the
"User") obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

1.  The above copyright notices and this permission notice (which
includes the disclaimer below) shall be included in all copies or
substantial portions of the Software.

2.  The name of a copyright holder shall not be used to endorse or
promote products derived from the Software without specific prior
written permission.

THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS
LICENSE.  NO USE OF THE SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER
THIS DISCLAIMER.  THE SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.  IN NO
EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  NO ASSURANCES ARE
PROVIDED BY THE COPYRIGHT HOLDERS THAT THE SOFTWARE DOES NOT INFRINGE
THE PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS OF ANY OTHER ENTITY.
EACH COPYRIGHT HOLDER DISCLAIMS ANY LIABILITY TO THE USER FOR CLAIMS
BROUGHT BY ANY OTHER ENTITY BASED ON INFRINGEMENT OF INTELLECTUAL
PROPERTY RIGHTS OR OTHERWISE.  AS A CONDITION TO EXERCISING THE RIGHTS
GRANTED HEREUNDER, EACH USER HEREBY ASSUMES SOLE RESPONSIBILITY TO SECURE
ANY OTHER INTELLECTUAL PROPERTY RIGHTS NEEDED, IF ANY.  THE SOFTWARE
IS NOT FAULT-TOLERANT AND IS NOT INTENDED FOR USE IN MISSION-CRITICAL
SYSTEMS, SUCH AS THOSE USED IN THE OPERATION OF NUCLEAR FACILITIES,
AIRCRAFT NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL
SYSTEMS, DIRECT LIFE SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH
THE FAILURE OF THE SOFTWARE OR SYSTEM COULD LEAD DIRECTLY TO DEATH,
PERSONAL INJURY, OR SEVERE PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH
RISK ACTIVITIES").  THE COPYRIGHT HOLDERS SPECIFICALLY DISCLAIM ANY
EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR HIGH RISK ACTIVITIES.

end of j2kaudio license

some documentation in j2kaudio.h
*/

#include "j2kaudio.h"

//default constructor
j2kaudio::j2kaudio(){
	version = "200708011100";
	jas_init();
	
	wavdata = NULL;
	samples = 0;
	wavarraysize = 0;

	sizeofframespersecond = sizeof(framespersecond);
	sizeofjp2datasize = sizeof(jp2datasize);
	sizeofsample = sizeof(sample);
	
	channels = 2;
	samplerate = 44100;
	bitspersample = sizeofsample*8;
	framespersecond = 1;
	frame = 0;
	encodingrate = 0.1;
	
	jp2data = NULL;
	jp2arraysize = 0;
		
	debugoutput = false;
	
	infile = NULL;
	outfile = NULL;

	numcmpts = 1;
	image = NULL;
	outputstream = NULL;
	inputstream = NULL;
	matrixdata = NULL;
	rate = new char[32];
}//end j2kaudio

j2kaudio::~j2kaudio(){
	jas_matrix_destroy(matrixdata);
	jas_image_destroy(image);
	jas_stream_close(outputstream);
	jas_stream_close(inputstream);	
}//end ~j2kaudio

string j2kaudio::getversion(){
	return version;
}//end getversion

void j2kaudio::setencodingrate(float er){
	encodingrate = er;
}//end setencodingrate





//function reads in an n-channel by s-sample audio 2d matrix
bool j2kaudio::readwav(const char* filename){
	if(debugoutput) cout << "Reading wav frame: " << frame << endl;
	if(frame == 0){
	
		//attempt to load the wav file using the external wav code in wav_in.h
		infile = new WAV_IN(filename);
		samplerate = (unsigned int)infile->get_sample_rate_hz();
		bitspersample = (unsigned char)infile->get_bits_per_sample();
		channels = (unsigned char)infile->get_num_channels();
		
		if(bitspersample != sizeofsample*8 && bitspersample != 24 && sizeofsample*8 != 32){
			cerr << "Error: wave file precision: " << (unsigned int)bitspersample << " != j2kaudio precision: " << sizeofsample*8 << endl;
			cerr << "Error: try using other j2kaudio version or change typedef sampletype\n";
			exit(1);
		}//end if
		
		//check for CD quality stereo audio as a development starting point
		/*if(samplerate != 44100 || bitspersample != 16 || channels != 2){
			cerr << "Error: SampleRate: " << samplerate << " BitsPerSample: " << (unsigned int)bitspersample << " AudioChannels: " << (unsigned int)channels << endl;
			cerr << "Error: The audio file is not standard CD quality audio, which is 44100 Hz, 16 bit, 2 channel audio." << endl;
			cerr.flush();
			exit(1);
		}//end if*/
		
		//allocate memory for our n-channel audio matrix
		wavdata = new sampletype*[channels];
		maxsamples = samplerate/framespersecond;
		wavarraysize = maxsamples*sizeofsample;
		if(debugoutput) cout << "about to malloc: " << wavarraysize << endl;
		for(unsigned char i=0; i < channels; i++)
			wavdata[i] = (sampletype*)malloc(wavarraysize);
		if(debugoutput) cout << "done malloc\n";
		
		imageheight = channels;
		jp2arraysize = maxsamples * imageheight * bitspersample/8;
	}//end if
	
	//read in some audio
	samples = 0;
	while(infile->more_data_available() && samples < maxsamples){
	 	//add the current sample to the correct audio channel object in the audio matrix
		for(unsigned char c=0; c < channels; c++){
			//samples in a multi-channel wav file are interleaved
			//read in the current sample
			sample = infile->read_current_input();
			wavdata[c][samples] = sample;
		}//end for
		samples++;
	}//end while
		
	imagewidth = (unsigned short)(samples);
	
	if(samples < maxsamples) return false;
	return true;
}//end readwav






//wav data to jp2 data function
//this function should receive an audio 2d array
//and output jp2 data to a buffer in j2a format
void j2kaudio::compress(){
	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general matrix to j2k steps
	//get data into jas_matrix_t objects
	//get matrix objects into image as a component
	//encode/compress the image into a stream object
	//write out the stream object to a jp2 file
	
	if(frame == 0){
		// Configure image component parameters
		if(debugoutput) cout << "Creating and configuring new image component parameters.\n";
		
		if(bitspersample != 32){
			numcmpts = 1;//1 component for anything under 29 bits/sample
			compp = new jas_image_cmptparm_t[1];
			compp[0].tlx = 0;
			compp[0].tly = 0;
			compp[0].hstep = 1;
			compp[0].vstep = 1;
			compp[0].width = imagewidth;
			compp[0].height = imageheight;			
			compp[0].sgnd = true;
			compp[0].prec = bitspersample;//single component only works up to 29 bits/sample 30-32 bits/sample results in assertion failure
		}//end if
		else{
			numcmpts = 3;//3 components for 32-bit float
			compp = new jas_image_cmptparm_t[3];
			compp[0].tlx = 0;
			compp[0].tly = 0;
			compp[0].hstep = 1;
			compp[0].vstep = 1;
			compp[0].width = imagewidth;
			compp[0].height = imageheight;			
			compp[0].sgnd = false;
			compp[0].prec = 1;//1 sign bit in 32-bit float
			compp[1].tlx = 0;
			compp[1].tly = 0;
			compp[1].hstep = 1;
			compp[1].vstep = 1;
			compp[1].width = imagewidth;
			compp[1].height = imageheight;
			compp[1].sgnd = false;
			compp[1].prec = 8;//8 exponent bits
			compp[2].tlx = 0;
			compp[2].tly = 0;
			compp[2].hstep = 1;
			compp[2].vstep = 1;
			compp[2].width = imagewidth;
			compp[2].height = imageheight;
			compp[2].sgnd = false;
			compp[2].prec = 23;//23 mantissa bits
		}//end else
		
		if(sizeofsample == 2)//16-bit signed
			compp[0].sgnd = true;
		else if(sizeofsample == 1)//8-bit unsigned
			compp[0].sgnd = false;
		if(debugoutput) cout << "compp[0].prec: " << compp[0].prec << endl;
		if(debugoutput) cout << "Just created and configured image component parameters.\n";
		
		// Create an image.
		if(debugoutput) cout << "Creating image.\n";
		image = jas_image_create(numcmpts, compp, JAS_CLRSPC_SGRAY);
		if(image == NULL){
			cerr << "Error: image is NULL.\n";
			cerr.flush();
			exit(1);
		}//end if
		if(debugoutput) cout << "Just created a " << image->brx_ << " x " << image->bry_ << " image with " << image->numcmpts_ << " components.\n";
			
		// Create a matrix
		if(debugoutput) cout << "Creating matrix data.\n";
		matrixdata = jas_matrix_create(imageheight, imagewidth);
		if(matrixdata == NULL){
			cerr << "Error: matrixdata is NULL.\n";
			cerr.flush();
			exit(1);
		}//end if
		if(debugoutput) cout << "Just created matrix data.\n";
		
		//allocate jp2data buffer for frame/image
		jp2data = new char[jp2arraysize];
		

	}//end if
	
	if(debugoutput) cout << "Encoding frame: " << frame << endl;

	//cycle through image components
	sampletype csample;
	sampletype signbit;
	sampletype exponentbits;
	sampletype mantissabits;
	for(unsigned short cc=0; cc < numcmpts; cc++){
		// Populate matrix with data
		if(debugoutput) cout << "Populating matrixdata.\n";
		for(unsigned char c = 0; c < imageheight; c++){
			for(unsigned short s = 0; s < imagewidth; s++){
				csample = wavdata[c][s];
				if(bitspersample == 32){
					if(debugoutput&&c==0&&s==10)cout << "csample " << s << ": " << (*(float*)&csample) << endl;
					if(cc == 0){//32-bit sign bit
						signbit = csample >> 31 & 0x00000001;//right-shift by 31 bits, leaving sign bit in bit 0, keep 1 bit
						if(debugoutput&&c==0&&s==10)cout << "signbit: " << signbit << " csample: " << csample << endl;
						csample = signbit;
					}//end else
					else if(cc == 1){//32-bit 8 exponent bits
						exponentbits = csample >> 23 & 0x000000FF;//right-shift by 23 bits, leaving exponent in bits 0-8, keep 8 bits
						if(debugoutput&&c==0&&s==10)cout << "exponentbits: " << exponentbits << endl;
						csample = exponentbits;
					}//end else
					else if(cc == 2){//32-bit 23 mantissa bits
						mantissabits = csample & 0x007FFFFF;//no shift, mantissa is in bits 0-23, keep 23 bits
						if(debugoutput&&c==0&&s==10)cout << "mantissabits: " << mantissabits << endl;
						csample = mantissabits;
					}//end else
				}//end if 32-bit
				jas_matrix_set(matrixdata, c, s, csample);
				if(debugoutput&&c==0&&s==10)cout << "after insert csample: " << csample << endl;
				if(csample != jas_matrix_get(matrixdata, c, s) && csample != 0){
					cerr << "Error: data not inserted into matrix properly.\n";
					exit(1);
				}//end if
			}//end for
		}//end for
		if(debugoutput) cout << "Just populated matrixdata.\n";
		
		// Write matrix data to image
		if(debugoutput) cout << "Writing matrixdata component to image.\n";
		if(jas_image_writecmpt(image,//image
			cc,//component number
			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			imagewidth,//jas_image_coord_t width
			imageheight,//jas_image_coord_t height
  			matrixdata)){//jas_matrix_t *data
			cerr << "Error: matrixdata not inserted into image object.\n";
			exit(1);
		}//end if
		if(debugoutput) cout << "Just wrote matrixdata component to image.\n";
	}//end for cc
	
	// Open a buffer as a stream.
	if(!(outputstream = jas_stream_memopen(jp2data, jp2arraysize))){//previously fopen
		cerr << "Error: Could not open stream for writing.\n";
		exit(1);
	}//end if
	
	// Perform JPEG 2000 wavelet compression/encoding
	if(debugoutput) cout << "Encoding the image object to a stream object.\n";
	if(debugoutput) cout << "Number of image components: " << jas_image_numcmpts(image) << endl;
	if(debugoutput) cout << "Maximum image components: " << image->maxcmpts_ << endl;
	if(image->cmpts_ == NULL)
		if(debugoutput) cout << "Per-component imformation is NULL.\n";
	if(image->cmprof_ == NULL)
		if(debugoutput) cout << "cmprof_ is NULL.\n";
	if(debugoutput) cout << "image->clrspc_: " << image->clrspc_ << endl;
	string encoderopts;
	sprintf(rate,"%f",encodingrate);
	if(encodingrate > 0 && encodingrate < 1){//==1lossless, <1 = lossy
		encoderopts = "rate=";
		encoderopts.append(rate);
		encoderopts.append("\n");
	}//end if
	else if(encodingrate != 1){
		cerr << "Error: invalid encoding rate: " << encodingrate << endl;
		exit(1);
	}//end else
	encoderopts.append("mode=int\ncblkwidth=256\ncblkheight=8 ");//these options 735,4, originally 1024,4 may help tune audio quality
	encoderopts[encoderopts.size()-1] = '\0';
	char* encopts = new char[encoderopts.size()];
	for(int i=0; i<encoderopts.size(); i++)
		encopts[i] = encoderopts[i];
	if(debugoutput) cout << "using encopts: " << encopts << endl;
	if(debugoutput) cout.flush();
	if(jas_image_encode(image, outputstream, jas_image_strtofmt("jp2"),encopts)){
		cerr << "Error: could not encode image to stream.\n";
		exit(1);
	}//end if
	if(debugoutput) cout << "Just encoded the image object to a stream object.\n";

	jp2datasize = outputstream->rwcnt_;
	if(debugoutput) cout << "outputstream->rwcnt_: " << outputstream->rwcnt_ << endl;
	
	//clean up
	jas_stream_close(outputstream);
}//end compress






void j2kaudio::writej2a(const char* filename){
	if(debugoutput) cout << "Writing j2a frame: " << frame << endl;
	if(frame == 0){
		output.open(filename,ios::trunc|ios::binary);
		output.write((char*)&framespersecond,sizeofframespersecond);
	}//end if
	
	output.write((char*)&jp2datasize,sizeofjp2datasize);
	output.write(jp2data,jp2datasize);
	
	frame++;
}//end writej2a

void j2kaudio::closej2a(){
	output.close();
}//end closej2a






bool j2kaudio::readj2a(const char* filename){
	if(debugoutput) cout << "Reading j2a frame: " << frame << endl;
	if(frame == 0){
		input.open(filename,ios::binary);
		input.read((char*)&framespersecond,sizeofframespersecond);
		
		wavarraysize = 10000000;
		jp2arraysize = 10000000;
		jp2data = (char*)malloc(jp2arraysize);
	}//end if

	jp2datasize=0;
	input.read((char*)&jp2datasize,sizeofjp2datasize);
	input.read(jp2data,jp2datasize);
	
	if(input.good()) return true;
	input.close();
	return false;
}//end readj2a








//jp2 data to wav data function
//this function should receive jp2 data from a buffer in j2a format
//and output an audio 2d array
void j2kaudio::decompress(){
	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general j2k to matrix steps
	//read in the jp2 file to a stream object
	//decode/decompress the stream object into an image object
	//check colorspace for GRAYscale
	//get image components into matrix objects
	//get data out of jas_matrix_t objects

	if(debugoutput) cout << "Decoding frame: " << frame << endl;

		// Open a buffer as a stream.
		if(debugoutput) cout << "jp2arraysize: " << jp2arraysize << endl;cout.flush();
		//if(debugoutput) for(unsigned int s=0; s < jp2arraysize; s++)cout << "jp2data[" << s << "]: " << jp2data[s] << endl;cout.flush();
		//jp2data = (char*)realloc(jp2data,jp2arraysize*100);
		if(debugoutput) cout << "about to memopen\n";cout.flush();
		if(!(inputstream = jas_stream_memopen(jp2data, jp2arraysize))){//previously fopen
			cerr << "Error: Could not open stream for reading.\n";
			exit(1);
		}//end if
		if(debugoutput) cout << "just did memopen\n";cout.flush();	
	
	if(debugoutput) cout << "Decoding the stream object to an image object.\n";
	/* read an image from a stream in a specified format. */
	if(!(image = jas_image_decode(inputstream, jas_image_strtofmt("jp2"),NULL))){
		cerr << "Error: could not decode image from stream.\n";
		exit(1);
	}//end if
	if(debugoutput) cout << "Just decoded the image object from a stream object.\n";
	
	//check the image to make sure it is in the grayscale colorspace
	if(jas_image_clrspc(image) != JAS_CLRSPC_SGRAY){
		cerr << "Error: image is not grayscale.\n";
		exit(1);
	}//end if
	
	//need to check for 16bit signed here instead of assuming
	if(frame==0){
		bitspersample = image->cmpts_[0]->prec_;
		if(bitspersample == 1)bitspersample = 32;
		for(int nc = 0; nc < jas_image_numcmpts(image); nc++)
			cout << "precision: " << image->cmpts_[nc]->prec_ << endl;
		if(bitspersample != sizeofsample*8){
			cerr << "Error: j2a file precision: " << (unsigned int)bitspersample << " != j2kaudio precision: " << sizeofsample*8 << endl;
			cerr << "Error: try using other j2kaudio version or change typedef sampletype\n";
			exit(1);
		}//end if
	}//end if
	
	numcmpts = jas_image_numcmpts(image);
	imagewidth = jas_image_width(image);
	imageheight = jas_image_height(image);
	if(debugoutput) cout << "Decoding a " << numcmpts << " component image that is " << imagewidth << " x " << imageheight << endl;
	
	if(frame == 0){
		if(debugoutput) cout << "Creating matrix data.\n";
		matrixdata = jas_matrix_create(imageheight, imagewidth);
		if(matrixdata == NULL){
			cerr << "Error: matrixdata is NULL.\n";
			cerr.flush();
			exit(1);
		}//end if
		if(debugoutput) cout << "Just created matrix data.\n";
	}//end if
	
	//cycle through image components
	sampletype csample;
	sampletype signbit;
	sampletype exponentbits;
	sampletype mantissabits;
	for(unsigned short cc=0; cc < numcmpts; cc++){
		if(debugoutput) cout << "Reading matrixdata components from image.\n";
		// Read a rectangular region of an image component.
		if(jas_image_readcmpt(image,//image
			cc,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			imagewidth,//jas_image_coord_t width
			imageheight,//jas_image_coord_t height
  			matrixdata)){//jas_matrix_t *data
			cerr << "Error: matrixdata not inserted into image object.\n";
			exit(1);
		}//end if
		if(debugoutput) cout << "Just read matrixdata components from image.\n";
		if(debugoutput) cout << "image width: " << imagewidth << endl;
		if(debugoutput) cout << "image height: " << imageheight << endl;
		
		if(debugoutput) cout << "Populating wavdata.\n";
		if(frame==0){
			channels=imageheight;
			samplerate = framespersecond*imagewidth;
		}//end if
		if(wavdata == NULL){
			wavdata = new sampletype*[channels];
			for(unsigned short c=0; c < channels; c++){
				wavdata[c] = (sampletype*)malloc(wavarraysize);
			}//end for
		}//end if
		samples = imagewidth;
		if(debugoutput) cout << " samples: " << samples << endl;cout.flush();
	
		for(int c=0; c < channels; c++){
			for(int s = 0; s < imagewidth; s++){
				csample = jas_matrix_get(matrixdata, c, s);
				if(debugoutput&&c==0&&s==10)cout << "csample " << s << ": " << csample << endl;
				if(bitspersample == 32){
					if(cc == 0){
						signbit = (csample << 31 & 0x80000000);//left-shift sign bit back to original location
						wavdata[c][s] = signbit;
						if(debugoutput&&c==0&&s==10)cout << "signbit: " << (signbit >> 31 & 0x1) << endl;
					}//end if
					else if(cc==1){
						exponentbits = (csample << 23 & 0x7F800000);//left-shift exponent bits back to original location
						wavdata[c][s] |= exponentbits;
						if(debugoutput&&c==0&&s==10)cout << "exponentbits: " << (exponentbits >> 23 & 0xFF) << endl;
					}//end else
					else if(cc==2){
						mantissabits = (csample & 0x007FFFFF);//mantissa is already in original location
						wavdata[c][s] |= mantissabits;
						if(debugoutput&&c==0&&s==10)cout << "mantissabits: " << (mantissabits & 0x007FFFFF) << endl;
					}//end else
				}//end if
				else
					wavdata[c][s] = csample; 
			}//end for
		}//end for
		if(debugoutput) cout << "Just populated wavdata.\n";
	}//end for
	
	jas_stream_close(inputstream);
}//end decompress








void j2kaudio::writewav(const char* filename){
	if(debugoutput) cout << "Writing wav frame: " << frame << endl;
	if(frame == 0){
		outfile = new WAV_OUT(samplerate, bitspersample, channels);
		
		//check for CD quality stereo audio as a development starting point
		/*if(samplerate != 44100 || bitspersample != 16 || channels != 2){
			cerr << "Error: SampleRate: " << samplerate << " BitsPerSample: " << (unsigned int)bitspersample << " AudioChannels: " << (unsigned int)channels << endl;
			cerr << "Error: The audio file is not standard CD quality audio, which is 44100 Hz, 16 bit, 2 channel audio." << endl;
			cerr.flush();
			exit(1);
		}//end if*/
	}//end if
	
	for(unsigned long long s = 0; s < samples; s++){
		for(unsigned char c = 0; c < channels; c++){
			sample = wavdata[c][s];
			outfile->write_current_output(sample);
		}//end for
	}//end for
	
	frame++;
}//end writewav



void j2kaudio::closewav(const char* filename){
	outfile->save_wave_file(filename);
}//end closewav



























