/* j2kaudio.h

j2kaudio License 200707231230

Copyright (c) 2007 Gregory Alan Hildstrom

All rights reserved.

Permission is hereby granted, free of charge, to any person (the
"User") obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

1.  The above copyright notices and this permission notice (which
includes the disclaimer below) shall be included in all copies or
substantial portions of the Software.

2.  The name of a copyright holder shall not be used to endorse or
promote products derived from the Software without specific prior
written permission.

THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS
LICENSE.  NO USE OF THE SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER
THIS DISCLAIMER.  THE SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.  IN NO
EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  NO ASSURANCES ARE
PROVIDED BY THE COPYRIGHT HOLDERS THAT THE SOFTWARE DOES NOT INFRINGE
THE PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS OF ANY OTHER ENTITY.
EACH COPYRIGHT HOLDER DISCLAIMS ANY LIABILITY TO THE USER FOR CLAIMS
BROUGHT BY ANY OTHER ENTITY BASED ON INFRINGEMENT OF INTELLECTUAL
PROPERTY RIGHTS OR OTHERWISE.  AS A CONDITION TO EXERCISING THE RIGHTS
GRANTED HEREUNDER, EACH USER HEREBY ASSUMES SOLE RESPONSIBILITY TO SECURE
ANY OTHER INTELLECTUAL PROPERTY RIGHTS NEEDED, IF ANY.  THE SOFTWARE
IS NOT FAULT-TOLERANT AND IS NOT INTENDED FOR USE IN MISSION-CRITICAL
SYSTEMS, SUCH AS THOSE USED IN THE OPERATION OF NUCLEAR FACILITIES,
AIRCRAFT NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL
SYSTEMS, DIRECT LIFE SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH
THE FAILURE OF THE SOFTWARE OR SYSTEM COULD LEAD DIRECTLY TO DEATH,
PERSONAL INJURY, OR SEVERE PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH
RISK ACTIVITIES").  THE COPYRIGHT HOLDERS SPECIFICALLY DISCLAIM ANY
EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR HIGH RISK ACTIVITIES.

end of j2kaudio license


Objective: compress stereo wav data using numerous jpeg 2000 jp2 image files as frames
initial benchmarking and testing done with CD quality 44100Hz 16bit stereo audio

j2kaudio extension and format: .j2a
sample rate: 44100 default
number of audio channels: image height: 2 default
channel layout for multimedia apps:
	row 0: left (mono)
	row 1: right (stereo)
	row 2: subwoofer (2.1)
	row 3: center (3.1)
	row 4: surround left (5.1)
	row 5: surround right (5.1)
	row 6: back (6.1) or rear left (7.1)
	row 7: rear right (7.1)
frames per second: 1 default
sample rate / frames per second: image width: default 44100 / 1 = 44100
bits per sample: 16 default 1-component grayscale images for 16-bit audio

j2a file byte layout:
unsigned char fps
unsigned int samplerate
unsigned char channels
unsigned char precision
unsigned int frames
unsigned int arbitrary header size
arbitrary header data
unsigned int bytes
jp2 image
bytes
image
bytes
image
...
until EOF

changelog
20070723 - finished updating j2kaudio and documenting comparison test
20070712 - started working on cleaner class implementation and revised j2a format using jp2 frames
20031106 - stopped development; basic theory working, but sound quality sucked; thought about frames
20030901 - tested various audio matrix layouts for image compression
20030701 - started development during downtime

*/

#ifndef J2KAUDIO_H
#define J2KAUDIO_H

#include <iostream>
#include <string>
#include <fstream>
#include <jasper/jasper.h>
#include "j2kaudiotype.h"
#include "wav_in.h"
#include "wav_out.h"
using namespace std;

class j2kaudio{
public:
	j2kaudio();				//default constructor
	~j2kaudio();				//default destructor
	string getversion();			//returns j2kaudio version
	void setencodingrate(float);		//sets encoding rate for JPEG 2000
	
	bool readwav(const char*);		//reads 1 frame of wav file into memory
	void compress();			//compresses wav file data into jp2 image in memory
	void writej2a(const char*);		//writes jp2 image to a file in j2a format
	void closej2a();			//finishes writing j2a and closes file
	
	bool readj2a(const char*);		//reads jp2 image/frame from a file in j2a format
	void decompress();			//decompresses jp2 image into wav data frame in memory
	void writewav(const char*);		//writes wav data frame to a wav file
	void closewav(const char*);		//finishes writing wav and closes file
private:
	string version;				//j2kaudio version
	sampletype** wavdata;			//2d array/buffer to hold wav data frame samples wide x channels high
	unsigned int samples;			//number of samples/channel in frame
	unsigned int maxsamples;		//number of max sample/channel in frame
	unsigned int wavarraysize;		//size of wav buffer (bytes) == max frame size
	sampletype sample;			//holds 1 sample of audio data
	unsigned char channels;			//number of channels
	unsigned int samplerate;		//sample rate per channel
	unsigned char bitspersample;		//bits per sample
	unsigned int framespersecond;		//number of audio frames per second
	unsigned int frame;			//current frame number
	float encodingrate;			//encoding rate for JPEG 2000 compression
	unsigned short imagewidth;		//image width of each frame = samplerate/framespersecond
	unsigned short imageheight;		//image height of each frame is number of channels
	
	char* jp2data;				//buffer to hold jp2 image encoded data frame
	unsigned int jp2arraysize;		//size of jp2data buffer
	unsigned int jp2datasize;		//actual bytes in buffer
	
	WAV_IN* infile;				//pointer to the wav input class
	WAV_OUT* outfile;			//pointer to the wav output class
	
	unsigned short sizeofframespersecond;	//size of variable, to prevent recalculation in loops
	unsigned short sizeofjp2datasize;	//size of variable, to prevent recalculation in loops 
	unsigned short sizeofsample;		//size of variable, to prevent recalculation in loops
	
	bool debugoutput;			//debugging output flag
	
	int numcmpts;				//number of image components
	jas_image_t * image;			//the resultant image
	jas_image_cmptparm_t* compp;		//component parameters
	jas_stream_t * outputstream;		//jp2 stream	
	jas_stream_t * inputstream;		//jp2 stream
	jas_matrix_t * matrixdata;		//component matrix data
	char* rate;				//string version of encodingrate
	
	ofstream output;			//output file stream
	ifstream input;				//input file stream
};//end class



#endif



